﻿// **********************************************************************************************************************************
// Example top-level C# code for interfacing with the Surrey Sensors Ltd. 64-channel 14-bit Pressure Scanner.
// Paul Nathan 26-FEB-2023.
// Copyright(c) 2023 Surrey Sensors Ltd. 

// By using this code, you are agreeing to the terms of the Surrey Sensors Ltd. end-user licence agreement, available at:
// https://www.surreysensors.com/wp-content/uploads/2021/08/SD-EULA-100821.pdf
// **********************************************************************************************************************************


using System;
using System.IO.Ports;

public class Example
{
    // User-configurable parameters for this example
    static int N_samples = 10;
    static double Data_Rate_Hz = 10.0;

    // Instance and constructor of a pressure scanner PS0
    static Pressure_Scanner PS0 = new Pressure_Scanner();

    static void Main()
    {
        // This example console application demonstrates the basics of pressure array data retrieval.
        //
        //     A. First list the available COM ports
        //     B. User then enters the port name associated with the device
        //     C. Initialize the desired COM port
        //
        // The example code then does the following:
        //
        //     1. Retrieve the device's serial number
        //     2. Set the device's data rate to 10 Hz
        //     3. Read back the device's data period after setting
        //     4. Check system power-on status
        //     5. Power on if required
        //     6. Update system status to confirm Power_On state and display some basic info
        //     7. Get 10 samples (1s worth), displaying only the pressure array data (formatted as an 8x8 table to 1 dp), followed by the Checksum ok flag
        //     8. Power off.
        //
        // Note that steps 1...6 are not required in order to run step 7 (data acquisition), they are just provided as example operations.
        // The system automatically powers on when plugged in.
        // It is recommended to check the system status beforehand to ensure there are no faults.
        // Step 8 is also optional for power saving or long idle time while remaining plugged in.


        // A.
        List_COM_Ports();

        // B.
        Console.WriteLine("Enter port name of the pressure scanner:");
        string portName = Console.ReadLine().ToUpper();
        Console.WriteLine("");

        // C.
        PS0.Initialize_Serial(portName);

        // 1.
        UInt32 Ser_Num = PS0.Get_Device_Ser_Num();
        Console.WriteLine("Device serial number:       {0}", Ser_Num);

        // 2.
        float Data_Period_us = (float)(1.0E6 / Data_Rate_Hz);
        PS0.Set_Data_Period(Data_Period_us);

        // 3.
        float Data_Period_returned = PS0.Get_Data_Period_us();
        double Data_Rate_returned = 1.0E6/ Data_Period_returned;
        Console.WriteLine("Data rate:                  {0} Hz", Data_Rate_returned);

        // 4.
        PS0.Get_System_Status();
        if (!PS0.System_Status.Power_On)
        {
            // 5.
            Console.WriteLine("System is powering up....");
            PS0.Power_On();
        }

        // 6.
        PS0.Get_System_Status();
        Console.WriteLine("System power on:            {0}", PS0.System_Status.Power_On);
        Console.WriteLine("Number of sensors detected: {0}", PS0.System_Status.N_sensors_detected);
        Console.WriteLine("Number of faulty sensors:   {0}", PS0.System_Status.N_Sensors_faulty);
        Console.WriteLine("");

        // 7.
        Console.WriteLine("Collecting {0} samples....", N_samples);
        PS0.Get_Samples(N_samples);

        Console.WriteLine("");
        for (int i = 0; i < N_samples; i++)
        {
            for (int j = 0; j < 8; j++)
            {
                for (int k = 0; k < 8; k++)
                {
                    Console.Write(PS0.Sample[i].Pressure[(8 * j) + k].ToString("F2") + "    ");
                }
                Console.WriteLine("");
            }
            Console.WriteLine("Checksum ok:                {0}", PS0.Sample[i].CRC_ok);
            Console.WriteLine("");
        }

        // 8.
        PS0.Power_Off();
        PS0.Get_System_Status();
        Console.WriteLine("System power on:            {0}", PS0.System_Status.Power_On);


        // Final clean-up before exiting permanently (optional)
        PS0.Serial.Dispose();

        Console.WriteLine("");
        Console.WriteLine("Press any key to exit");
        Console.ReadLine();
    }


    static void List_COM_Ports()
    {
        Console.WriteLine("Available COM ports:");

        string[] s = SerialPort.GetPortNames();

        for (int i = 0; i < s.Length; i++)
        {
            Console.WriteLine(s[i]);
        }
        Console.WriteLine("");
    }
}