﻿// **********************************************************************************************************************************
// C# code for interfacing with the Surrey Sensors Ltd. 64-channel 14-bit Pressure Scanner.
// Paul Nathan 26-FEB-2023.
// Copyright(c) 2023 Surrey Sensors Ltd. 

// By using this code, you are agreeing to the terms of the Surrey Sensors Ltd. end-user licence agreement, available at:
// https://www.surreysensors.com/wp-content/uploads/2021/08/SD-EULA-100821.pdf
// **********************************************************************************************************************************


using System;
using System.IO.Ports;
using System.Text;
using System.Threading;


public class Pressure_Scanner
{
    // Command string constants
    private class Command
    {
        public const string Power_On = "@P";
        public const string Power_Off = "@p";
        public const string Start_Streaming = "@D";
        public const string Stop_Streaming = "@d";
        public const string Enable_HW_Trig = "@H";
        public const string Disable_HW_Trig = "@h";
        public const string Set_Data_Period = "@F";
        public const string Get_Data_Period = "@f";
        public const string Get_Ser_Num = "@N";
        public const string Temporary_Autozero = "@z";
        public const string Run_Self_Test = "@S";
        public const string Get_Status_Bytes = "@s";
        public const string Get_Sensor_Params = "@W";
        public const string Override_N_Active = "@O";
        public const string System_Reset = "@R";
    }


    // Data packet and sample array
    public class Data
    {
        public float[] Pressure = new float[64];
        public float T_thermistor;
        public float P_atmos;
        public float Rel_Hum;
        public float T_board;
        public float[] accel = new float[3];
        public float[] gyro = new float[3];
        public byte[] Sensor_status_bytes = new byte[8];
        public byte clock_warn;
        public ushort CRC;
        public bool CRC_ok;
    };
    public Data[] Sample;


    // System status parsed from status bytes
    public class SysStatus
    {
        public bool Power_On;
        public bool MCU_Checksum_Ok;
        public bool Thermistor_Ok;
        public bool IMU_detected;
        public bool Accelerometer_Ok;
        public bool Gyroscope_Ok;
        public bool Env_sensor_detected;
        public bool[] Pressure_Sensor_Present = new bool[64];
        public bool[] Pressure_Sensor_Good = new bool[64];
        public int N_sensors_detected, N_sensors_good, N_Sensors_faulty;
    };
    public SysStatus System_Status = new SysStatus();


    // Serial port object
    public SerialPort Serial;

    // System endianness
    private readonly bool BigEndian = !BitConverter.IsLittleEndian;

    // Serial port parameters
    private const int baudRate = 12000000;
    private const int timeout_ms = 5000;

    // Serial data frame character
    private const char FrameChar = '#';

    // Other arrays (status bytes, sensor params)
    public byte[] System_status_bytes = new byte[17];
    public float[] P_min = new float[64];
    public float[] P_max = new float[64];
    public float[] P_offset = new float[64];


    // #############################################################################################################################
    public void Initialize_Serial(string portName)
    {
        Serial = new SerialPort(portName, baudRate, Parity.None, 8, StopBits.One);

        Serial.Encoding = Encoding.ASCII;
        Serial.DiscardNull = false;

        Serial.Handshake = Handshake.None;
        Serial.DtrEnable = false;
        Serial.RtsEnable = false;

        Serial.ReadTimeout = timeout_ms;
        Serial.WriteTimeout = timeout_ms;
    }


    private void Send_Command(string Command)
    {
        Serial.Open();                                           // Open serial port
        Serial.Write(Command);                                   // Send command string
        Serial.Close();                                          // Close serial port
    }


    public void Power_On()
    {
        Send_Command(Command.Power_On);
        Thread.Sleep(3000);
    }

    public void Power_Off()
    {
        Send_Command(Command.Power_Off);
    }

    public void Enable_HW_Trig()
    {
        Send_Command(Command.Enable_HW_Trig);
    }

    public void Disable_HW_Trig()
    {
        Send_Command(Command.Disable_HW_Trig);
    }

    public void System_Reset()
    {
        Send_Command(Command.System_Reset);
    }


    public UInt32 Get_Device_Ser_Num()
    {
        const int n = sizeof(UInt32);
        byte[] buffer = new byte[n];

        Serial.Open();                                           // Open serial port

        Serial.Write(Command.Get_Ser_Num);                       // Send command string

        while (Serial.BytesToRead == 0) ;                        // Wait until byte(s) received from device
        Serial.Read(buffer, 0, n);                               // Read back requested data

        Serial.Close();                                          // Close serial port

        if (BigEndian) Array.Reverse(buffer);                    // Convert to computer's endianness from device's little-endianness
        UInt32 Ser_Num = BitConverter.ToUInt32(buffer, 0);       // Cast byte array to uint32

        return Ser_Num;
    }


    public float Get_Data_Period_us()
    {
        const int n = sizeof(float);
        byte[] buffer = new byte[n];

        Serial.Open();                                           // Open serial port

        Serial.Write(Command.Get_Data_Period);                   // Send command string

        while (Serial.BytesToRead == 0) ;                        // Wait until byte(s) received from device
        Serial.Read(buffer, 0, n);                               // Read back requested data

        Serial.Close();                                          // Close serial port

        if (BigEndian) Array.Reverse(buffer);                    // Convert to computer's endianness from device's little-endianness
        float Data_Period_us = BitConverter.ToSingle(buffer, 0); // Cast byte array to float

        return Data_Period_us;
    }


    public void Set_Data_Period(float Data_Period_us)
    {
        const int n = sizeof(float);

        byte[] buffer = BitConverter.GetBytes(Data_Period_us);   // Cast float to byte array
        if (BigEndian) Array.Reverse(buffer);                    // Convert to device's little-endianness from computer's endianness

        Serial.Open();                                           // Open serial port

        Serial.Write(Command.Set_Data_Period);                   // Send command string
        Serial.Write(buffer, 0, n);                              // Send data bytes

        Serial.Close();                                          // Close serial port
    }


    public void Run_Self_Test()
    {
        const int Len = 17;
        byte[] buffer = new byte[Len];

        Serial.Open();                                           // Open serial port

        Serial.Write(Command.Run_Self_Test);                     // Send command string

        while (Serial.BytesToRead == 0) ;                        // Wait until byte(s) received from device
        Serial.Read(buffer, 0, Len);                             // Read back requested data

        Serial.Close();                                          // Close serial port

        Array.Copy(buffer, System_status_bytes, Len);            // Copy direct to array (since simply bytes)
        Parse_System_Status_Bytes();                             // Parse status bytes 
    }


    public void Get_System_Status()
    {
        const int Len = 17;
        byte[] buffer = new byte[Len];

        Serial.Open();                                           // Open serial port

        Serial.Write(Command.Get_Status_Bytes);                  // Send command string

        while (Serial.BytesToRead == 0) ;                        // Wait until byte(s) received from device
        Serial.Read(buffer, 0, Len);                             // Read back requested data

        Serial.Close();                                          // Close serial port

        Array.Copy(buffer, System_status_bytes, Len);            // Copy direct to array (since simply bytes)
        Parse_System_Status_Bytes();                             // Parse status bytes 
    }


    public void Get_Samples(int N_samples)
    {
        Sample = new Data[N_samples];

        const int Len = 308;
        const int Lenm1 = Len - 1;
        const int Lenm2 = Len - 2;
        byte[] buffer = new byte[Len];

        Serial.Open();                                           // Open serial port

        Serial.Write(Command.Start_Streaming);                   // Send start command string
        while (Serial.BytesToRead == 0) ;                        // Wait until byte(s) received from device

        int i = 0;
        while (i < N_samples)
        {
            Serial.Read(buffer, 0, 1);                           // Read first byte, check it's the correct frame character
            if (buffer[0] == FrameChar)
            {
                Serial.Read(buffer, 1, Lenm1);                   // Read remaining bytes and parse data
                Parse_Serial_Data(buffer, Lenm2, i);
                ++i;                                             // Increment sample number
            }
            else
            {
                // Frame error, keep stepping through buffer until alignment
                // this error should not occurr in normal operation
            }

        }

        Serial.Write(Command.Stop_Streaming);                    // Send stop command string

        Serial.Close();                                          // Close serial port
    }


    private void Parse_Serial_Data(byte[] buffer, int Lenm2, int k)  // k is the sample number
    {
        const int n = sizeof(float);

        // Constructor for Sample class array element k
        Sample[k] = new Data();

        // 64x pressures
        int index = 1;
        for (int i = 0; i < 64; i++)
        {
            if (BigEndian) Array.Reverse(buffer, index, n);
            Sample[k].Pressure[i] = BitConverter.ToSingle(buffer, index);
            index += n;
        }

        // Thermistor
        index = 257;
        if (BigEndian) Array.Reverse(buffer, index, n);
        Sample[k].T_thermistor = BitConverter.ToSingle(buffer, index);

        // P_atmos
        index = 261;
        if (BigEndian) Array.Reverse(buffer, index, n);
        Sample[k].P_atmos = BitConverter.ToSingle(buffer, index);

        // Relative humidity
        index = 265;
        if (BigEndian) Array.Reverse(buffer, index, n);
        Sample[k].Rel_Hum = BitConverter.ToSingle(buffer, index);

        // T_board
        index = 269;
        if (BigEndian) Array.Reverse(buffer, index, n);
        Sample[k].T_board = BitConverter.ToSingle(buffer, index);

        // ax, ay, az
        index = 273;
        if (BigEndian) Array.Reverse(buffer, index, n);
        Sample[k].accel[0] = BitConverter.ToSingle(buffer, index);
        index = 277;
        if (BigEndian) Array.Reverse(buffer, index, n);
        Sample[k].accel[1] = BitConverter.ToSingle(buffer, index);
        index = 281;
        if (BigEndian) Array.Reverse(buffer, index, n);
        Sample[k].accel[2] = BitConverter.ToSingle(buffer, index);

        // wx, wy, wz
        index = 285;
        if (BigEndian) Array.Reverse(buffer, index, n);
        Sample[k].gyro[0] = BitConverter.ToSingle(buffer, index);
        index = 289;
        if (BigEndian) Array.Reverse(buffer, index, n);
        Sample[k].gyro[1] = BitConverter.ToSingle(buffer, index);
        index = 293;
        if (BigEndian) Array.Reverse(buffer, index, n);
        Sample[k].gyro[2] = BitConverter.ToSingle(buffer, index);

        // 8x bank status bytes
        for (int i = 0; i < 8; i++)
        {
            index = 297 + i;
            Sample[k].Sensor_status_bytes[i] = buffer[index];
        }

        Sample[k].clock_warn = buffer[305];

        //CRC16 check
        index = 306;
        if (BigEndian) Array.Reverse(buffer, index, sizeof(UInt16));
        Sample[k].CRC = BitConverter.ToUInt16(buffer, index);

        UInt16 CRC_Calc = CRC16.Compute_CRC16(buffer, Lenm2, CRC16.CRC16_init);
        if (CRC_Calc == Sample[k].CRC)
        {
            Sample[k].CRC_ok = true;
        }
    }


    public void Temporary_Auto_Zero()
    {
        const int n = sizeof(float);
        const int Len = 64 * n;
        byte[] buffer = new byte[Len];

        Serial.Open();                                           // Open serial port

        Serial.Write(Command.Temporary_Autozero);                // Send command string

        Thread.Sleep(10000);                                     // Wait 10s for operation to complete

        while (Serial.BytesToRead == 0) ;                        // Wait until byte(s) received from device
        Serial.Read(buffer, 0, Len);                             // Read back requested data

        Serial.Close();                                          // Close serial port

        // Parse 64x pressure offsets
        int index = 0;
        for (int i = 0; i < 64; i++)
        {
            if (BigEndian) Array.Reverse(buffer, index, n);
            P_offset[i] = BitConverter.ToSingle(buffer, index);
            index += n;
        }
    }


    public void Get_Sensor_Params()
    {
        const int n = sizeof(float);
        const int Len = 3 * 64 * n;
        byte[] buffer = new byte[Len];

        Serial.Open();                                           // Open serial port

        Serial.Write(Command.Get_Sensor_Params);                 // Send command string

        while (Serial.BytesToRead == 0) ;                        // Wait until byte(s) received from device
        Serial.Read(buffer, 0, Len);                             // Read back requested data

        Serial.Close();                                          // Close serial port

        // Parse 64x triplets of P_min, P_max, P_offset
        int index = 0;
        for (int i = 0; i < 64; i++)
        {
            if (BigEndian) Array.Reverse(buffer, index, n);
            P_min[i] = BitConverter.ToSingle(buffer, index);
            index += 4;

            if (BigEndian) Array.Reverse(buffer, index, n);
            P_max[i] = BitConverter.ToSingle(buffer, index);
            index += 4;

            if (BigEndian) Array.Reverse(buffer, index, n);
            P_offset[i] = BitConverter.ToSingle(buffer, index);
            index += 4;
        }
    }


    public void Override_N_Active(byte N_active)
    {
        const int n = sizeof(byte);

        byte[] buffer = new byte[n];                
        buffer[0] = N_active;

        Serial.Open();                                           // Open serial port

        Serial.Write(Command.Override_N_Active);                 // Send command string
        Serial.Write(buffer, 0, n);                              // Send data bytes

        Serial.Close();                                          // Close serial port
    }


    private void Parse_System_Status_Bytes()
    {
        // byte index 0
        System_Status.Power_On = (System_status_bytes[0] & 0x01) > 0;
        System_Status.MCU_Checksum_Ok = (System_status_bytes[0] & 0x02) > 0;
        System_Status.Thermistor_Ok = (System_status_bytes[0] & 0x04) > 0;
        System_Status.IMU_detected = (System_status_bytes[0] & 0x08) > 0;
        System_Status.Accelerometer_Ok = (System_status_bytes[0] & 0x10) > 0;
        System_Status.Gyroscope_Ok = (System_status_bytes[0] & 0x20) > 0;
        System_Status.Env_sensor_detected = (System_status_bytes[0] & 0x40) > 0;

        // byte indices  1...16
        int j = 0, k = 0, l;
        for (int i = 1; i < 9; i++)
        {
            System_Status.Pressure_Sensor_Present[j++] = (System_status_bytes[i] & 0x01) > 0;
            System_Status.Pressure_Sensor_Present[j++] = (System_status_bytes[i] & 0x02) > 0;
            System_Status.Pressure_Sensor_Present[j++] = (System_status_bytes[i] & 0x04) > 0;
            System_Status.Pressure_Sensor_Present[j++] = (System_status_bytes[i] & 0x08) > 0;
            System_Status.Pressure_Sensor_Present[j++] = (System_status_bytes[i] & 0x10) > 0;
            System_Status.Pressure_Sensor_Present[j++] = (System_status_bytes[i] & 0x20) > 0;
            System_Status.Pressure_Sensor_Present[j++] = (System_status_bytes[i] & 0x40) > 0;
            System_Status.Pressure_Sensor_Present[j++] = (System_status_bytes[i] & 0x80) > 0;

            l = 8 + i;
            System_Status.Pressure_Sensor_Good[k++] = (System_status_bytes[l] & 0x01) > 0;
            System_Status.Pressure_Sensor_Good[k++] = (System_status_bytes[l] & 0x02) > 0;
            System_Status.Pressure_Sensor_Good[k++] = (System_status_bytes[l] & 0x04) > 0;
            System_Status.Pressure_Sensor_Good[k++] = (System_status_bytes[l] & 0x08) > 0;
            System_Status.Pressure_Sensor_Good[k++] = (System_status_bytes[l] & 0x10) > 0;
            System_Status.Pressure_Sensor_Good[k++] = (System_status_bytes[l] & 0x20) > 0;
            System_Status.Pressure_Sensor_Good[k++] = (System_status_bytes[l] & 0x40) > 0;
            System_Status.Pressure_Sensor_Good[k++] = (System_status_bytes[l] & 0x80) > 0;
        }

        // Get number of detected and good sensors
        System_Status.N_sensors_detected = 0;
        System_Status.N_sensors_good = 0;
        for (int i = 0; i < 64; i++)
        {
            if (System_Status.Pressure_Sensor_Present[i])
            {
                ++System_Status.N_sensors_detected;
                if (System_Status.Pressure_Sensor_Good[i]) ++System_Status.N_sensors_good;
            }
        }
        System_Status.N_Sensors_faulty = System_Status.N_sensors_detected - System_Status.N_sensors_good;
    }
}